;;
;;	keyboard module
;;
;;	external interface, matrix keyboard, 12 keys, telephone type.
;;	peripheral resources, port b, tris b, interrupt on change.
;;	internal interface, initialise, interrupt service routine, get.
;;

p_kbc		equ	portb		; keyboard column outputs
t_kbc		equ	trisb
b_kbc0		equ	0	; kb pin 3	1 4 7 *
b_kbc1		equ	1	; kb pin 1	2 5 8 0
b_kbc2		equ	2	; kb pin 5	3 6 9 #
m_kbc		equ	(1<<b_kbc0)|(1<<b_kbc1)|(1<<b_kbc2)

p_kbr		equ	portb		; keyboard row inputs
t_kbr		equ	trisb
b_kbr0		equ	4	; kb pin 4	* 0 #
b_kbr1		equ	5	; kb pin 6	7 8 9
b_kbr2		equ	6	; kb pin 7	4 5 6
b_kbr3		equ	7	; kb pin 2	1 2 3
m_kbr		equ	(1<<b_kbr0)|(1<<b_kbr1)|(1<<b_kbr2)|(1<<b_kbr3)
	
	cblock
		kb_is:3		; new key mask 
		kb_was:3	; old key mask
				; high four bits of each used, wasteful
		kb_key		; key pressed, or 0 for none
	endc
	
kb_initialise
	;; drive all column pins to ground
	bcf	p_kbr,b_kbc0
	bcf	p_kbr,b_kbc1
	bcf	p_kbr,b_kbc2
	;; set all column pins as output
	bsf	status,rp0	; { bank 1
	bcf	t_kbr,b_kbc0
	bcf	t_kbr,b_kbc1
	bcf	t_kbr,b_kbc2
	;; set all row pins to be inputs (for interrupt on change)
	bsf	t_kbc,b_kbr0
	bsf	t_kbc,b_kbr1
	bsf	t_kbc,b_kbr2
	bsf	t_kbc,b_kbr3
	;; enable weak pull-up on rows (in particular)
	bcf	option_reg,not_rbpu	; enable weak pull-up for portb
	bcf	status,rp0	; } bank 1
	;; clear previous state
	clrf	kb_key
	movlw	0xff
	movwf	kb_was
	movwf	kb_was+1
	movwf	kb_was+2
	bcf	flag_kb
	;; enable interrupt on change on columns (in particular)
	bsf	intcon,rbie
	return

	;; fetch all into a buffer, one bit per key, waste four bits
kb_read	macro			; output in kb_is
	;; read rows on column 0, clearing the interrupt
	bsf	status,rp0	; { bank 1
	bsf	t_kbr,b_kbc1
	bsf	t_kbr,b_kbc2
	bcf	status,rp0	; } bank 1

	movf	p_kbr,w
	movwf	kb_is

	;; read rows on column 1
	bsf	status,rp0	; { bank 1
	bsf	t_kbr,b_kbc0
	bcf	t_kbr,b_kbc1
	bcf	status,rp0	; } bank 1
	
	movf	p_kbr,w
	movwf	kb_is+1

	;; read rows on column 2
	bsf	status,rp0	; { bank 1
	bsf	t_kbr,b_kbc1
	bcf	t_kbr,b_kbc2
	bcf	status,rp0	; } bank 1
	
	movf	p_kbr,w
	movwf	kb_is+2

	;; restore column drivers to ground for next time
	bsf	status,rp0	; { bank 1
	bcf	t_kbr,b_kbc0
	bcf	t_kbr,b_kbc1
	bcf	status,rp0	; } bank 1
	
	movf	p_kbr,w		; read port again to clear 
				; interrupt caused by restoring column drivers
	
	endm

	;; test the down event mask for a row and enqueue keys
kb_row	macro	col,key
	rlf	kb_was+col,f
	movlw	key
	btfsc	status,c
	call	kb_put
	endm

	;; process each column, detecting down events
kb_col	macro	col,bit7,bit6,bit5,bit4
	
	;; detect down events, kb_is bit zero with kb_was bit one
	movf	kb_was+col,w
	xorwf	kb_is+col,w
	andwf	kb_was+col,f

	;;       down  hold   up
	;; was = 1111  1110  1110
	;; is  = 1110  1110  1111
	;; xor = 0001  0000  0001
	;; and = 0001  0000  0000
	;; each high bit in kb_was now represents buttons that closed

	kb_row	col,bit7
	kb_row	col,bit6
	kb_row	col,bit5
	kb_row	col,bit4
	endm
	
	;; copy new buffer to old buffer
kb_save	macro
	movf	kb_is,w
	movwf	kb_was
	movf	kb_is+1,w
	movwf	kb_was+1
	movf	kb_is+2,w
	movwf	kb_was+2
	endm
	
kb_isr
	nop			; so i can get here in gpsim
	kb_read			; fetch all into a buffer, one bit per key
	kb_col	0,'1','4','7','*'
	kb_col	1,'2','5','8','0'
	kb_col	2,'3','6','9','#'
	kb_save			; copy new buffer to old buffer
	return

kb_put
	movwf	kb_key		; temporary one key deep buffer
	bsf	flag_kb		; wake mainline with new key
	return

kb_get				; ( -- key|0 )
	pushf	kb_key
	clrf	kb_key
	bcf	flag_kb
	bcf	intcon,rbie	; temporarily disable port change interrupt
	call	kb_isr		; see note below
	bsf	intcon,rbie	; enable port change interrupt
	return

	;; release of a key sometimes will not trigger the port change
	;; interrupt, as it may have been released before the last read
	;; of port before return from interrupt.
	
	;; test rig for keyboard unit testing
kb_unit_test_scan	macro
	pushl	pos_data
	call	lcd_move

	pushf	kb_key
	call	lcd_put_hex

	call	lcd_put_space

	pushf	kb_is
	call	lcd_put_hex
	
	pushf	kb_is+1
	call	lcd_put_hex
	
	pushf	kb_is+2
	call	lcd_put_hex
	
	endm
